
class Node(object):
  def __init__(self):
    self.value = None 
    self.error = None
    self.samples = None
    self.feature_threshold = None
    self.feature_index = None
    self.left_child = None
    self.right_child = None
    self.height = None
    self.nid = None
    self.left_nid = None
    self.right_nid = None

  def __str__(self):
    if self.left_child and self.right_child:
      return "[NODE] Height: %s, FeatureIndex: %s, Samples: %d" % (self.height, self.feature_index, self.samples)
    else:
      return "[LEAF] Height: %s, Samples: %s" % (self.height, self.samples)






import numpy as np
from cuda_random_decisiontree_small import RandomDecisionTreeSmall
from util import timer, get_best_dtype, dtype_to_ctype, mk_kernel, mk_tex_kernel
from pycuda import gpuarray

class RandomForestClassifier(object):
  """A random forest classifier.

    A random forest is a meta estimator that fits a number of classifical
    decision trees on various sub-samples of the dataset and use averaging
    to improve the predictive accuracy and control over-fitting.
    
    Usage:
    See RandomForestClassifier.fit
  """ 
  COMPT_THREADS_PER_BLOCK = 64 
  RESHUFFLE_THREADS_PER_BLOCK = 64 
  
  def __compact_labels(self, target):
    def check_is_compacted(x):
      return x.size == int(np.max(x)) + 1 and int(np.min(x)) == 0
    def convert_to_dict(x):
      d = {}
      for i, val in enumerate(x):
        d[val] = i
      return d

    self.compt_table = np.unique(target)
    self.compt_table.sort()        
    if not check_is_compacted(self.compt_table):
      trans_table = convert_to_dict(self.compt_table)
      for i, val in enumerate(target):
        target[i] = trans_table[val]

  def fit(self, samples, target, n_trees = 10, min_samples_split = 1, max_features = None, max_depth = None):
    """Construce multiple trees in the forest.

    Parameters
    ----------
    samples:numpy.array of shape = [n_samples, n_features]
            The training input samples.

    target: numpy.array of shape = [n_samples]
            The training input labels.

    n_trees : integer, optional (default=10)
        The number of trees in the forest.

    max_features : int or None, optional (default="log2(n_features)")
        The number of features to consider when looking for the best split:
          - If None, then `max_features=log2(n_features)`.

    max_depth : integer or None, optional (default=None)
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_spl samples.

    min_samples_split : integer, optional (default=2)
        The minimum number of samples required to split an internal node.
        Note: this parameter is tree-specific.
    
    Returns
    -------
    None
    """
    assert isinstance(samples, np.ndarray)
    assert isinstance(target, np.ndarray)
    assert samples.size / samples[0].size == target.size

    target = target.copy()
    self.__compact_labels(target)
    self.n_labels = self.compt_table.size 

    self.dtype_indices = get_best_dtype(target.size)
    self.dtype_counts = self.dtype_indices
    self.dtype_labels = get_best_dtype(self.n_labels)
    self.dtype_samples = samples.dtype
   
    samples = np.require(np.transpose(samples), requirements = 'C')
    target = np.require(np.transpose(target), dtype = self.dtype_labels, requirements = 'C') 
    self.n_features = samples.shape[0]
    self.n_samples = target.size
    
    samples_gpu = gpuarray.to_gpu(samples)
    labels_gpu = gpuarray.to_gpu(target) 
    
    sorted_indices = np.empty((self.n_features, target.size), dtype = self.dtype_indices)
    
    for i,f in enumerate(samples):
      sort_idx = np.argsort(f)
      sorted_indices[i] = sort_idx  
 
    self.forest = [RandomDecisionTreeSmall(samples_gpu, labels_gpu, sorted_indices, self.compt_table, 
      self.dtype_labels,self.dtype_samples, self.dtype_indices, self.dtype_counts,
      self.n_features, self.n_samples, self.n_labels, self.COMPT_THREADS_PER_BLOCK,
      self.RESHUFFLE_THREADS_PER_BLOCK, max_features, max_depth, min_samples_split) for i in xrange(n_trees)]   
   
    for i, tree in enumerate(self.forest):
      with timer("Tree %s" % (i,)):
        tree.fit(samples, target)

  def predict(self, x):
    """Predict labels for giving samples.

    Parameters
    ----------
    x:numpy.array of shape = [n_samples, n_features]
            The predicting input samples.
    
    Returns
    -------
    y: Array of shape [n_samples].
        The predicted labels.
    """
    res = []
    for tree in self.forest:
      res.append(tree.gpu_predict(x))
    res = np.array(res)
    return np.array([np.argmax(np.bincount(res[:,i])) for i in xrange(res.shape[1])]) 




import numpy as np
from util import get_best_dtype
from pycuda import gpuarray

class BaseTree(object):
  def __init__(self):
    self.root = None
    self.max_depth = None

  def print_tree(self):
    def recursive_print(node):
      if node.left_child and node.right_child:
        print node
        recursive_print(node.left_child)
        recursive_print(node.right_child)
      else:
        print node
    assert self.root is not None
    recursive_print(self.root)

  def __predict(self, val):
    temp = self.root
    while True:
      if temp.left_child and temp.right_child:
        if val[temp.feature_index] < temp.feature_threshold:
          temp = temp.left_child
        else:
          temp = temp.right_child
      else:
          if hasattr(self, "compt_table"):
            return self.compt_table[temp.value]
          else:
            return temp.value

  def predict(self, inputs):
    res = []
    for val in inputs:
      res.append(self.__predict(val))
    return np.array(res)

  def __gpu_predict(self, val):
    idx = 0
    while True:
      threshold = self.threshold_array[idx]
      threshold_idx = self.feature_array[idx]
      left_idx = self.left_child_array[idx]
      right_idx = self.right_child_array[idx]

      if left_idx != 0 and right_idx != 0: #Means it has children
        if val[threshold_idx] < threshold:
          idx = left_idx
        else:
          idx = right_idx
      else:
        return self.value_array[idx]

  def gpu_predict(self, inputs):
    inputs = np.require(inputs.copy(), requirements = "C")
    n_predict = inputs.shape[0]    
    predict_gpu = gpuarray.to_gpu(inputs)
    left_child_gpu = gpuarray.to_gpu(self.left_child_array)
    right_child_gpu = gpuarray.to_gpu(self.right_child_array)
    threshold_gpu = gpuarray.to_gpu(self.threshold_array)
    value_gpu = gpuarray.to_gpu(self.value_array)
    feature_gpu = gpuarray.to_gpu(self.feature_array)
    predict_res_gpu = gpuarray.zeros(n_predict, dtype=self.dtype_labels)

    self.predict_kernel.prepared_call(
                  (n_predict, 1),
                  (1, 1, 1),
                  left_child_gpu.ptr,
                  right_child_gpu.ptr,
                  feature_gpu.ptr,
                  threshold_gpu.ptr,
                  value_gpu.ptr,
                  predict_gpu.ptr,
                  predict_res_gpu.ptr,
                  self.n_features,
                  self.n_nodes)
    
    if hasattr(self, "compt_table"):
      return np.array([self.compt_table[i] for i in predict_res_gpu.get()], dtype = self.compt_table.dtype)
    else:
      return predict_res_gpu.get()

  def decorate_nodes(self, samples, labels):
    """ In __construct function, the node just store the indices of the actual data, now decorate it with the actual data."""
    def recursive_decorate(node):
      if node.left_child and node.right_child:
        idx_tuple = node.feature_threshold
        node.feature_threshold = (float(samples[idx_tuple[0], idx_tuple[1]]) + samples[idx_tuple[0], idx_tuple[2]]) / 2
        recursive_decorate(node.left_child)
        recursive_decorate(node.right_child)
      else:
        idx = node.value
        node.value = labels[idx]
        
    assert self.root != None
    recursive_decorate(self.root)
  
  def __find_most_common_label(self, x):
    return np.argmax(np.bincount(x))

  def gpu_decorate_nodes(self, samples, labels):
    """ Decorate the nodes for GPU predicting """
    def recursive_decorate(node):
      if node.left_child and node.right_child:
        idx_tuple = node.feature_threshold
        node.feature_threshold = (float(samples[idx_tuple[0], idx_tuple[1]]) + samples[idx_tuple[0], idx_tuple[2]]) / 2
        
        self.left_child_array[node.nid] = node.left_nid
        self.right_child_array[node.nid] = node.right_nid
        self.threshold_array[node.nid] = node.feature_threshold
        self.feature_array[node.nid] = node.feature_index

        recursive_decorate(node.left_child)
        recursive_decorate(node.right_child)
      else:
        if isinstance(node.value, np.ndarray):
          res = labels[node.value]
          self.value_array[node.nid] = self.__find_most_common_label(res)
        else:
          idx = node.value
          node.value = labels[idx]
          self.value_array[node.nid] = node.value

    self.left_child_array = np.zeros(self.n_nodes, self.dtype_indices)
    self.right_child_array = np.zeros(self.n_nodes, self.dtype_indices)
    self.threshold_array = np.zeros(self.n_nodes, np.float32)
    self.value_array = np.zeros(self.n_nodes, self.dtype_labels) 
    self.feature_array = np.zeros(self.n_nodes, np.uint16)

    assert self.root != None
    recursive_decorate(self.root)



import pycuda.autoinit
import pycuda.driver as cuda
from pycuda import gpuarray
from sklearn import tree
import numpy as np
import math
from time import sleep
import datasource
from util import mk_kernel, mk_tex_kernel, timer, dtype_to_ctype, get_best_dtype, start_timer, end_timer
from node import Node
from cuda_random_base_tree import RandomBaseTree
from collections import deque
import time

class RandomDecisionTreeSmall(RandomBaseTree): 
  def __init__(self, samples_gpu, labels_gpu, sorted_indices, compt_table, dtype_labels, dtype_samples, 
      dtype_indices, dtype_counts, n_features, n_samples, n_labels, n_threads, n_shf_threads, max_features = None,
      max_depth = None, min_samples_split = None):
    self.root = None
    self.n_labels = n_labels
    self.max_depth = None
    self.stride = n_samples
    self.dtype_labels = dtype_labels
    self.dtype_samples = dtype_samples
    self.dtype_indices = dtype_indices
    self.dtype_counts = dtype_counts
    self.n_features = n_features
    self.COMPT_THREADS_PER_BLOCK = n_threads
    self.RESHUFFLE_THREADS_PER_BLOCK = n_shf_threads
    self.samples_gpu = samples_gpu
    self.labels_gpu = labels_gpu
    self.sorted_indices = sorted_indices
    self.compt_table = compt_table
    self.max_depth = max_depth
    self.max_features = max_features
    self.min_samples_split =  min_samples_split

  def __compile_kernels(self):
    ctype_indices = dtype_to_ctype(self.dtype_indices)
    ctype_labels = dtype_to_ctype(self.dtype_labels)
    ctype_counts = dtype_to_ctype(self.dtype_counts)
    ctype_samples = dtype_to_ctype(self.dtype_samples)
    n_labels = self.n_labels
    n_threads = self.COMPT_THREADS_PER_BLOCK
    n_shf_threads = self.RESHUFFLE_THREADS_PER_BLOCK
    
    self.fill_kernel = mk_kernel((ctype_indices,), "fill_table", "fill_table_si.cu") 
    self.scan_total_kernel = mk_kernel((n_threads, n_labels, ctype_labels, ctype_counts, ctype_indices), 
        "count_total", "scan_kernel_total_si.cu") 
    
    self.comput_total_kernel = mk_kernel((n_threads, n_labels, ctype_samples, ctype_labels, 
      ctype_counts, ctype_indices), "compute", "comput_kernel_total_rand.cu")
     
    self.scan_reshuffle_tex, tex_ref = mk_tex_kernel((ctype_indices, n_shf_threads), 
        "scan_reshuffle", "tex_mark", "pos_scan_reshuffle_si_c_tex.cu")   
    self.mark_table.bind_to_texref_ext(tex_ref) 
    
    self.comput_label_loop_kernel = mk_kernel((n_threads, n_labels, ctype_samples, 
      ctype_labels, ctype_counts, ctype_indices), "compute",  "comput_kernel_label_loop_si.cu") 
    
    self.comput_label_loop_rand_kernel = mk_kernel((n_threads, n_labels, ctype_samples, 
      ctype_labels, ctype_counts, ctype_indices), "compute",  "comput_kernel_label_loop_rand.cu") 
    
    self.find_min_kernel = mk_kernel((ctype_counts, 32), "find_min_imp", "find_min_gini.cu")
      
    self.predict_kernel = mk_kernel((ctype_indices, ctype_samples, ctype_labels), "predict", "predict.cu")
  
    self.scan_total_bfs = mk_kernel((32, n_labels, ctype_labels, ctype_counts, ctype_indices), "count_total", "scan_kernel_total_bfs.cu")
  
    self.comput_bfs = mk_kernel((32, n_labels, ctype_samples, ctype_labels, ctype_counts, ctype_indices), "compute", "comput_kernel_bfs.cu")
    
    self.fill_bfs = mk_kernel((ctype_indices,), "fill_table", "fill_table_bfs.cu")
    
    self.reshuffle_bfs = mk_kernel((ctype_indices, 32), "scan_reshuffle", "pos_scan_reshuffle_bfs.cu")
    
    if hasattr(self.fill_kernel, "is_prepared"):
      return
    
    self.fill_kernel.is_prepared = True
    self.fill_kernel.prepare("PiiPi")
    self.scan_reshuffle_tex.prepare("PPPiii") 
    self.scan_total_kernel.prepare("PPPi")
    self.comput_total_kernel.prepare("PPPPPPPPii")
    self.comput_label_loop_kernel.prepare("PPPPPPPii")
    self.comput_label_loop_rand_kernel.prepare("PPPPPPPPii")
    self.find_min_kernel.prepare("PPPi")
    self.predict_kernel.prepare("PPPPPPPii")
    self.scan_total_bfs.prepare("PPPPPPPi")
    self.comput_bfs.prepare("PPPPPPPPPPPPii")
    self.fill_bfs.prepare("PPPPPPPi")
    self.reshuffle_bfs.prepare("PPPPPPPii")

  def __allocate_gpuarrays(self):
    if self.max_features < 4:
      imp_size = 4
    else:
      imp_size = self.max_features
    self.impurity_left = gpuarray.empty(imp_size, dtype = np.float32)
    self.impurity_right = gpuarray.empty(self.max_features, dtype = np.float32)
    self.min_split = gpuarray.empty(self.max_features, dtype = self.dtype_counts)
    self.mark_table = gpuarray.empty(self.stride, dtype = np.uint8)
    self.label_total = gpuarray.empty(self.n_labels, self.dtype_indices)  
    self.subset_indices = gpuarray.empty(self.max_features, dtype = self.dtype_indices)
  
  def __release_gpuarrays(self):
    self.impurity_left = None
    self.impurity_right = None
    self.min_split = None
    self.mark_table = None
    self.label_total = None
    self.subset_indices = None
    self.sorted_indices_gpu = None
    self.sorted_indices_gpu_ = None
  
  def __bfs_construct(self):
    while len(self.queue):
      self.__bfs()

  def __bfs(self):
    idx_array = np.zeros(len(self.queue) * 2, dtype = self.dtype_indices)
    si_idx_array = np.zeros(len(self.queue), dtype = np.uint8) 
    subset_indices_array = np.zeros(len(self.queue) * self.max_features, dtype = self.dtype_indices)
    
    queue_size = len(self.queue)
    
    for i, node in enumerate(self.queue):
      idx_array[i * 2] = node.start_idx
      idx_array[i * 2 + 1] = node.stop_idx
      si_idx_array[i] = node.si_idx
      subset_indices_array[i * self.max_features : (i + 1) * self.max_features] = node.subset_indices  

    idx_array_gpu = gpuarray.to_gpu(idx_array)
    si_idx_array_gpu = gpuarray.to_gpu(si_idx_array)
    subset_indices_array_gpu = gpuarray.to_gpu(subset_indices_array)
    min_feature_idx_gpu = gpuarray.empty(queue_size, dtype = np.uint16)
    
    self.label_total = gpuarray.empty(queue_size * self.n_labels, dtype = self.dtype_counts)  
    self.impurity_left = gpuarray.empty(queue_size, dtype = np.float32)
    self.impurity_right = gpuarray.empty(queue_size, dtype = np.float32)
    self.min_split = gpuarray.empty(queue_size, dtype = self.dtype_indices)

    if len(self.mark_table.shape) == 1:
      self.mark_table = gpuarray.zeros((self.n_features, self.stride), dtype=np.uint8)
    
    self.scan_total_bfs.prepared_call(
            (queue_size, 1),
            (32, 1, 1),
            self.sorted_indices_gpu.ptr,
            self.sorted_indices_gpu_.ptr,
            self.labels_gpu.ptr,
            self.label_total.ptr,
            si_idx_array_gpu.ptr,
            idx_array_gpu.ptr,
            subset_indices_array_gpu.ptr,
            self.max_features) 
    
    self.comput_bfs.prepared_call(
          (queue_size, 1),
          (32, 1, 1),
          self.samples_gpu.ptr,
          self.labels_gpu.ptr,
          self.sorted_indices_gpu.ptr,
          self.sorted_indices_gpu_.ptr,
          idx_array_gpu.ptr,
          si_idx_array_gpu.ptr,
          self.label_total.ptr,
          subset_indices_array_gpu.ptr,
          self.impurity_left.ptr,
          self.impurity_right.ptr,
          self.min_split.ptr,
          min_feature_idx_gpu.ptr,
          self.max_features,
          self.stride)
    
    self.fill_bfs.prepared_call(
          (queue_size, 1),
          (32, 1, 1),
          self.sorted_indices_gpu.ptr,
          self.sorted_indices_gpu_.ptr,
          si_idx_array_gpu.ptr,
          min_feature_idx_gpu.ptr,
          idx_array_gpu.ptr,
          self.min_split.ptr,
          self.mark_table.ptr,
          self.stride)
   
    self.reshuffle_bfs.prepared_call(
          (queue_size, 16),
          (32, 1, 1),
          self.mark_table.ptr,
          si_idx_array_gpu.ptr,
          self.sorted_indices_gpu.ptr,
          self.sorted_indices_gpu_.ptr,
          idx_array_gpu.ptr,
          self.min_split.ptr,
          min_feature_idx_gpu.ptr,
          self.n_features,
          self.stride)
    
    begin_end_idx = idx_array_gpu.get()
    imp_left = self.impurity_left.get()
    imp_right = self.impurity_right.get()
    min_split = self.min_split.get()
    feature_idx = min_feature_idx_gpu.get()

    for i in xrange(queue_size):
      node = self.queue.popleft()
      node.feature_index = feature_idx[i]
      
      if node.si_idx == 1:
        si = self.sorted_indices_gpu
      else:
        si = self.sorted_indices_gpu_

      row = node.feature_index
      col = min_split[i]
      left_imp = imp_left[i]
      right_imp = imp_right[i]

      cuda.memcpy_dtoh(self.threshold_value_idx, si.ptr +  
          int(row * self.stride + col) * int(self.dtype_indices.itemsize))
      node.feature_threshold = (row, self.threshold_value_idx[0], self.threshold_value_idx[1])
     
      if left_imp + right_imp == 4.0:
        self.__record_leaf(node, node.start_idx, node.stop_idx - node.start_idx, si)
        continue
       
      left_node =  Node()
      right_node = Node()
      
      left_node.depth = node.depth + 1
      left_node.nid = self.n_nodes 
      node.left_nid = left_node.nid
      self.n_nodes += 1  
      
      right_node.depth = left_node.depth
      right_node.nid = self.n_nodes
      node.right_nid = right_node.nid
      self.n_nodes += 1
      si_id = 1 - node.si_idx
      
      if left_imp != 0.0:
        n_samples_left = col + 1 - node.start_idx 
        if n_samples_left < self.min_samples_split or (self.max_depth is not None and left_node.depth >= self.max_depth):
          self.__record_leaf(left_node, node.start_idx, n_samples_left, si)
        else:
          left_node.start_idx = node.start_idx
          left_node.stop_idx = col + 1
          left_node.si_idx = si_id
          left_node.subset_indices = self.get_indices()  
          self.queue.append(left_node)
      else:
        cuda.memcpy_dtoh(self.target_value_idx, si.ptr + int(node.start_idx * self.dtype_indices.itemsize))
        left_node.value = self.target_value_idx[0]

      if right_imp != 0.0:
        n_samples_right = node.stop_idx - col - 1
        if n_samples_right < self.min_samples_split or (self.max_depth is not None and right_node.depth >= self.max_depth):
          self.__record_leaf(right_node, col + 1, n_samples_right, si)
        else:
          right_node.start_idx = col + 1
          right_node.stop_idx = node.stop_idx
          right_node.si_idx = si_id
          right_node.subset_indices = self.get_indices()
          self.queue.append(right_node)
      else:
        cuda.memcpy_dtoh(self.target_value_idx, si.ptr + int((col + 1) * self.dtype_indices.itemsize))
        
        right_node.value = self.target_value_idx[0]   
      node.left_child = left_node
      node.right_child = right_node

  def fit(self, samples, target): 
    self.samples_itemsize = self.dtype_samples.itemsize
    self.labels_itemsize = self.dtype_labels.itemsize
    self.target_value_idx = np.zeros(1, self.dtype_indices)
    self.threshold_value_idx = np.zeros(2, self.dtype_indices)
    self.min_imp_info = np.zeros(4, dtype = np.float32)  
    self.queue = deque()
    
    if self.max_features is None:
      self.max_features = int(math.ceil(math.log(self.n_features, 2)))

    assert self.max_features > 0 and self.max_features <= self.n_features, "max_features must be between 0 and n_features." 
    self.__allocate_gpuarrays()
    self.__compile_kernels() 
    self.sorted_indices_gpu = gpuarray.to_gpu(self.sorted_indices)
    self.sorted_indices_gpu_ = self.sorted_indices_gpu.copy()
    
    self.sorted_indices_gpu.idx = 0
    self.sorted_indices_gpu_.idx = 1

    assert self.sorted_indices_gpu.strides[0] == target.size * self.sorted_indices_gpu.dtype.itemsize 
    assert self.samples_gpu.strides[0] == target.size * self.samples_gpu.dtype.itemsize   
    
    self.n_nodes = 0 
    self.root = self.__dfs_construct(1, 1.0, 0, target.size, self.sorted_indices_gpu, self.sorted_indices_gpu_, self.get_indices())  
    self.__bfs_construct() 
    self.__release_gpuarrays()
    self.gpu_decorate_nodes(samples, target)
 
  def __record_leaf(self, ret_node, start_idx, n_samples, si):
      """ Pick the indices to record on the leaf node. In decoration step, we'll choose the most common label """
      if n_samples < 3:
        cuda.memcpy_dtoh(self.target_value_idx, si.ptr + int(start_idx * self.dtype_indices.itemsize))
        ret_node.value = self.target_value_idx[0] 
      else:
        si_labels = np.empty(n_samples, dtype=self.dtype_indices)
        cuda.memcpy_dtoh(si_labels, si.ptr + int(start_idx * self.dtype_indices.itemsize))
        ret_node.value = si_labels


  def __dfs_construct(self, depth, error_rate, start_idx, stop_idx, si_gpu_in, si_gpu_out, subset_indices):
    def check_terminate():
      if error_rate == 0:
        return True
      else:
        return False 

    n_samples = stop_idx - start_idx
    indices_offset =  start_idx * self.dtype_indices.itemsize
 
    ret_node = Node()
    ret_node.error = error_rate
    ret_node.samples = n_samples 
    ret_node.height = depth 
    ret_node.nid = self.n_nodes
    ret_node.depth = depth

    self.n_nodes += 1

    if check_terminate():
      cuda.memcpy_dtoh(self.target_value_idx, si_gpu_in.ptr + int(start_idx * self.dtype_indices.itemsize))
      ret_node.value = self.target_value_idx[0] 
      return ret_node
    
    if n_samples < self.min_samples_split or (self.max_depth is not None and depth >= self.max_depth):
      self.__record_leaf(ret_node, start_idx, n_samples, si_gpu_in)
      return ret_node
    
    if n_samples <= 64:
      ret_node.start_idx = start_idx
      ret_node.stop_idx = stop_idx
      ret_node.si_idx = si_gpu_in.idx
      ret_node.subset_indices = subset_indices
      self.queue.append(ret_node)
      return ret_node 
    

    cuda.memcpy_htod(self.subset_indices.ptr, subset_indices)
    grid = (self.max_features, 1) 
    
    if n_samples >= self.COMPT_THREADS_PER_BLOCK:
      block = (self.COMPT_THREADS_PER_BLOCK, 1, 1)
    else:
      block = (32, 1, 1)
     
    self.scan_total_kernel.prepared_call(
                (1, 1),
                block,
                si_gpu_in.ptr + indices_offset,
                self.labels_gpu.ptr,
                self.label_total.ptr,
                n_samples)

    self.comput_label_loop_rand_kernel.prepared_call(
                grid,
                block,
                si_gpu_in.ptr + indices_offset,
                self.samples_gpu.ptr,
                self.labels_gpu.ptr,
                self.impurity_left.ptr,
                self.impurity_right.ptr,
                self.label_total.ptr,
                self.min_split.ptr,
                self.subset_indices.ptr,
                n_samples,
                self.stride)
    
    #self.comput_total_kernel.prepared_call(
    #            grid,
    #            block,
    #            si_gpu_in.ptr + indices_offset,
    #            self.samples_gpu.ptr,
    #            self.labels_gpu.ptr,
    #            self.impurity_left.ptr,
    #            self.impurity_right.ptr,
    #            self.label_total.ptr,
    #            self.min_split.ptr,
    #            self.subset_indices.ptr,
    #            n_samples,
    #            self.stride)

    subset_indices_left = self.get_indices()
    subset_indices_right = self.get_indices()
    
    self.find_min_kernel.prepared_call(
                (1, 1),
                (32, 1, 1),
                self.impurity_left.ptr,
                self.impurity_right.ptr,
                self.min_split.ptr,
                self.max_features)
    
    cuda.memcpy_dtoh(self.min_imp_info, self.impurity_left.ptr)
    min_right = self.min_imp_info[1] 
    min_left = self.min_imp_info[0] 
    
    if min_left + min_right == 4:
      self.__record_leaf(ret_node, start_idx, n_samples, si_gpu_in)
      return ret_node
     
    col = int(self.min_imp_info[2]) 
    row = int(self.min_imp_info[3])
    row = subset_indices[row]
    ret_node.feature_index = row
    
    cuda.memcpy_dtoh(self.threshold_value_idx, si_gpu_in.ptr + int(indices_offset) + 
        int(row * self.stride + col) * int(self.dtype_indices.itemsize))
    ret_node.feature_threshold = (row, self.threshold_value_idx[0], self.threshold_value_idx[1])
   
    self.fill_kernel.prepared_call(
                      (1, 1),
                      (1024, 1, 1),
                      si_gpu_in.ptr + row * self.stride * self.dtype_indices.itemsize + indices_offset, 
                      n_samples, 
                      col, 
                      self.mark_table.gpudata, 
                      self.stride)
       
    if n_samples >= self.RESHUFFLE_THREADS_PER_BLOCK:
      block = (self.RESHUFFLE_THREADS_PER_BLOCK, 1, 1)
    else:
      block = (32, 1, 1)

    self.scan_reshuffle_tex.prepared_call(
                      (self.n_features, 1),
                      block,
                      self.mark_table.ptr,
                      si_gpu_in.ptr + indices_offset,
                      si_gpu_out.ptr + indices_offset,
                      n_samples,
                      col,
                      self.stride) 

    ret_node.left_nid = self.n_nodes
    ret_node.left_child = self.__dfs_construct(depth + 1, min_left, 
        start_idx, start_idx + col + 1, si_gpu_out, si_gpu_in, subset_indices_left)
    
    ret_node.right_nid = self.n_nodes
    ret_node.right_child = self.__dfs_construct(depth + 1, min_right, 
        start_idx + col + 1, stop_idx, si_gpu_out, si_gpu_in, subset_indices_right)
    return ret_node 




"""
This Module is not used in the package.
"""

import pycuda.autoinit
import pycuda.driver as cuda
from pycuda import gpuarray
from sklearn import tree
import numpy as np
import math
from time import sleep
import datasource
from util import mk_kernel, mk_tex_kernel, timer, dtype_to_ctype, get_best_dtype
from node import Node
from cuda_base_tree import BaseTree

class DecisionTree(BaseTree): 
  COMPT_THREADS_PER_BLOCK = 64  #The number of threads do computation per block.
  RESHUFFLE_THREADS_PER_BLOCK = 64 

  def __init__(self):
    self.root = None
    self.compt_kernel_type = None
    self.num_labels = None
    self.label_count = None
    self.max_depth = None
    self.stride = None
    self.dtype_labels = None
    self.dtype_samples = None
    self.dtype_indices = None
    self.dtype_counts = None

  def fit(self, samples, target, max_depth = None):
    def compile_kernels():
      ctype_indices = dtype_to_ctype(self.dtype_indices)
      ctype_labels = dtype_to_ctype(self.dtype_labels)
      ctype_counts = dtype_to_ctype(self.dtype_counts)
      ctype_samples = dtype_to_ctype(self.dtype_samples)
      n_labels = self.num_labels
      n_threads = self.COMPT_THREADS_PER_BLOCK
      n_shf_threads = self.RESHUFFLE_THREADS_PER_BLOCK
      
      self.fill_kernel = mk_kernel((ctype_indices,), "fill_table", "fill_table_si.cu") 
      
      self.scan_reshuffle_kernel = mk_kernel((ctype_indices, n_shf_threads), 
          "scan_reshuffle", "pos_scan_reshuffle_si_c.cu")
      
      self.scan_total_kernel = mk_kernel((n_threads, n_labels, ctype_labels, ctype_counts, ctype_indices), 
          "count_total", "scan_kernel_total_si.cu") 
      
      self.comput_total_kernel, tex_ref = mk_tex_kernel((n_threads, n_labels, ctype_samples, ctype_labels, 
        ctype_counts, ctype_indices), "compute", "tex_label_total", "comput_kernel_total_si.cu")
      self.label_total.bind_to_texref_ext(tex_ref)
      
      self.scan_reshuffle_tex, tex_ref = mk_tex_kernel((ctype_indices, n_shf_threads), 
          "scan_reshuffle", "tex_mark", "pos_scan_reshuffle_si_c_tex.cu")   
      self.mark_table.bind_to_texref_ext(tex_ref)
      
      self.comput_label_loop_kernel, tex_ref = mk_tex_kernel((n_threads, n_labels, ctype_samples, 
        ctype_labels, ctype_counts, ctype_indices), "compute", "tex_label_total", "comput_kernel_label_loop_si.cu") 
      self.label_total.bind_to_texref_ext(tex_ref)
    
      self.find_min_kernel = mk_kernel((ctype_counts, 32), "find_min_imp", "find_min_gini.cu")
      
      self.predict_kernel = mk_kernel((ctype_indices, ctype_samples, ctype_labels), "predict", "predict.cu")

      """ Use prepare to improve speed """
      self.fill_kernel.prepare("PiiPi")
      self.scan_reshuffle_kernel.prepare("PPPiiiii")
      self.scan_reshuffle_tex.prepare("PPPiii") 
      self.scan_total_kernel.prepare("PPPi")
      self.comput_total_kernel.prepare("PPPPPPPii")
      self.comput_label_loop_kernel.prepare("PPPPPPPii")
      self.find_min_kernel.prepare("PPPi")
      self.predict_kernel.prepare("PPPPPPPii")

    def allocate_gpuarrays():
      """ Pre-allocate the GPU memory, don't allocate everytime in __construct"""
      self.impurity_left = gpuarray.empty(self.n_features, dtype = np.float32)
      self.impurity_right = gpuarray.empty(self.n_features, dtype = np.float32)
      self.min_split = gpuarray.empty(self.n_features, dtype = self.dtype_counts)
      self.mark_table = gpuarray.empty(target.size, dtype = np.uint8)
      self.label_count = gpuarray.empty((self.COMPT_THREADS_PER_BLOCK + 1) * self.num_labels * self.n_features, 
          dtype = self.dtype_counts)  
      self.label_total = gpuarray.empty(self.num_labels, self.dtype_indices)  
    
    def sort_arrays():
      sorted_indices = np.empty((self.n_features, target.size), dtype = self.dtype_indices)
      with timer("argsort"):
        for i,f in enumerate(samples):
          sort_idx = np.argsort(f)
          sorted_indices[i] = sort_idx  
      self.sorted_indices_gpu = gpuarray.to_gpu(sorted_indices)
      self.sorted_indices_gpu_ = self.sorted_indices_gpu.copy()
      self.samples_gpu = gpuarray.to_gpu(samples)
      self.labels_gpu = gpuarray.to_gpu(target)

    assert isinstance(samples, np.ndarray)
    assert isinstance(target, np.ndarray)
    assert samples.size / samples[0].size == target.size
    
    self.max_depth = max_depth
    self.num_labels = int(np.max(target)) + 1    
    self.stride = target.size

    self.dtype_indices = get_best_dtype(target.size)
    self.dtype_counts = self.dtype_indices
    self.dtype_labels = get_best_dtype(self.num_labels)
    self.dtype_samples = samples.dtype
     
    samples = np.require(np.transpose(samples), requirements = 'C')
    target = np.require(np.transpose(target), dtype = self.dtype_labels, requirements = 'C') 
    
    self.samples_itemsize = self.dtype_samples.itemsize
    self.labels_itemsize = self.dtype_labels.itemsize
    self.target_value_idx = np.zeros(1, self.dtype_indices)
    self.threshold_value_idx = np.zeros(2, self.dtype_indices)
    self.min_imp_info = np.zeros(4, dtype = np.float32)

    self.n_features = samples.shape[0]
    self.n_nodes = 0

    allocate_gpuarrays()
    compile_kernels()
    sort_arrays() 

    assert self.sorted_indices_gpu.strides[0] == target.size * self.sorted_indices_gpu.dtype.itemsize 
    assert self.samples_gpu.strides[0] == target.size * self.samples_gpu.dtype.itemsize   
    self.root = self.__construct(1, 1.0, 0, target.size, self.sorted_indices_gpu, self.sorted_indices_gpu_) 
    self.gpu_decorate_nodes(samples, target) 


  def __construct(self, depth, error_rate, start_idx, stop_idx, si_gpu_in, si_gpu_out):
    def check_terminate():
      if error_rate == 0 or (self.max_depth is not None and depth > self.max_depth):
        return True
      else:
        return False 
    
    n_samples = stop_idx - start_idx
    indices_offset =  start_idx * self.dtype_indices.itemsize
 
    ret_node = Node()
    ret_node.error = error_rate
    ret_node.samples = n_samples 
    ret_node.height = depth 
    ret_node.nid = self.n_nodes

    self.n_nodes += 1

    if check_terminate():
      cuda.memcpy_dtoh(self.target_value_idx, si_gpu_in.ptr + int(start_idx * self.dtype_indices.itemsize))
      ret_node.value = self.target_value_idx[0] 
      return ret_node
  
    grid = (self.n_features, 1) 
    block = (self.COMPT_THREADS_PER_BLOCK, 1, 1)

    self.scan_total_kernel.prepared_call(
                (1, 1),
                block,
                si_gpu_in.ptr + indices_offset,
                self.labels_gpu.ptr,
                self.label_total.ptr,
                n_samples)

    self.comput_label_loop_kernel.prepared_call(
                grid,
                block,
                si_gpu_in.ptr + indices_offset,
                self.samples_gpu.ptr,
                self.labels_gpu.ptr,
                self.impurity_left.ptr,
                self.impurity_right.ptr,
                self.label_total.ptr,
                self.min_split.ptr,
                n_samples,
                self.stride)
    
    self.find_min_kernel.prepared_call(
                (1, 1),
                (32, 1, 1),
                self.impurity_left.ptr,
                self.impurity_right.ptr,
                self.min_split.ptr,
                self.n_features)
    
    cuda.memcpy_dtoh(self.min_imp_info, self.impurity_left.ptr)
    min_right = self.min_imp_info[1] 
    min_left = self.min_imp_info[0] 
    col = int(self.min_imp_info[2]) 
    row = int(self.min_imp_info[3])
    ret_node.feature_index = row
    
    #self.comput_total_kernel.prepared_call(
    #            grid,
    #            block,
    #            si_gpu_in.ptr + indices_offset,
    #            self.samples_gpu.ptr,
    #            self.labels_gpu.ptr,
    #            self.impurity_left.ptr,
    #            self.impurity_right.ptr,
    #            self.label_total.ptr,
    #            self.min_split.ptr,
    #            n_samples,
    #            self.stride)

    if min_right + min_left == 4:
      print "######## depth : %d, n_samples: %d, row: %d, col: %d, start: %d, stop: %d" % (depth, n_samples, row, col, start_idx, stop_idx)
      return ret_node
    
    cuda.memcpy_dtoh(self.threshold_value_idx, si_gpu_in.ptr + int(indices_offset) + 
        int(row * self.stride + col) * int(self.dtype_indices.itemsize))
    ret_node.feature_threshold = (row, self.threshold_value_idx[0], self.threshold_value_idx[1])
    
    self.fill_kernel.prepared_call(
                      (1, 1),
                      (1024, 1, 1),
                      si_gpu_in.ptr + row * self.stride * self.dtype_indices.itemsize + indices_offset, 
                      n_samples, 
                      col, 
                      self.mark_table.gpudata, 
                      self.stride
                      )
    
    block = (self.RESHUFFLE_THREADS_PER_BLOCK, 1, 1)
    
    self.scan_reshuffle_tex.prepared_call(
                      grid,
                      block,
                      self.mark_table.ptr,
                      si_gpu_in.ptr + indices_offset,
                      si_gpu_out.ptr + indices_offset,
                      n_samples,
                      col,
                      self.stride) 
  

    ret_node.left_nid = self.n_nodes 
    ret_node.left_child = self.__construct(depth + 1, min_left, 
        start_idx, start_idx + col + 1, si_gpu_out, si_gpu_in)
    
    ret_node.right_nid = self.n_nodes
    ret_node.right_child = self.__construct(depth + 1, min_right, 
        start_idx + col + 1, stop_idx, si_gpu_out, si_gpu_in)
    return ret_node 



if __name__ == "__main__":
  x_train, y_train = datasource.load_data("digits") 
  with timer("Cuda"):
    d = DecisionTree()  
    d.fit(x_train, y_train, max_depth = None)
    d.print_tree()
  
  with timer("Predict"):
    #d.gpu_predict(x_train)
    #print y_train
    print np.allclose(d.gpu_predict(x_train), y_train)




import numpy as np
from cuda_base_tree import BaseTree
import random

class RandomBaseTree(BaseTree):
  def __init__(self):
    self.max_features = None
    self.n_features = None
    self.dtype_indices = None

  def get_indices(self):
    if self.max_features < self.n_features / 2:
      return np.array(random.sample(xrange(self.n_features), self.max_features), dtype=self.dtype_indices)
    else:
      return np.random.permutation(self.n_features)[:self.max_features].astype(self.dtype_indices)



#!/usr/bin/python
from setuptools import setup, find_packages
import cudatree

setup(
    name = "cudatree",
    long_decsription='''
CudaTree
========
Building Decision Tree on Cuda.
''',
    version = cudatree.__version__,
    description = "Building decison tree on Cuda",
    author = ["Alex Rubinsteyn", "Yisheng Liao"],
    author_email = ["alexr@cs.nyu.edu", "yl1912@nyu.edu"],
    packages = find_packages() + ['cudatree.test'],
    package_dir = {'cudatree.test' : './test'},
    package_data = {'cudatree' : ['cuda_kernels/*.cu']},
    requires = [
        'numpy',
        'sklearn',
        'pycuda'
      ],
    classifiers=['Development Status :: 3 - Alpha',
                  'Topic :: Software Development :: Libraries',
                  'License :: OSI Approved :: BSD License',
                  'Intended Audience :: Developers',
                  'Programming Language :: Python :: 2.7',
                ])


